"""
Tools for working with traceback.
"""
import inspect
import traceback
import types


def fmt_stack_summary(summary: traceback.StackSummary) -> str:
    return "".join(
        list(traceback.StackSummary.from_list(summary).format())).strip()

def get_as_str(err: Exception) -> str | None:
    s = None
    tb = err.__traceback__
    if tb:
        summary = traceback.extract_tb(tb)
        s = fmt_stack_summary(summary)
    return s

def set(
    err: Exception,
    skip_frames: int = 0,
    ignore_existing: bool = False,
):
    """
    Creates traceback for an err.

    If ``ignore_existing`` is true, and err already has a traceback, it will
    be overwritten. Otherwise for errs with traceback nothing will be done.

    Original err is not affected, modified err is returned. If nothing is done,
    the same err is returned without copying.

    Argument ``skip_frames`` defines how many frames to skip. This function
    or any nested function frames are automatically skipped.
    """
    if err.__traceback__ is not None and ignore_existing:
        err.__traceback__ = None

    # skip 2 frames - this call and this function call
    prev_tb: types.TracebackType | None = new(skip_frames + 2)

    err.__traceback__ = prev_tb

def new(skip_frames: int = 0) -> types.TracebackType | None:
    current_frame = inspect.currentframe()
    if current_frame is None:
        raise ValueError("unavailable to retrieve current frame")
    # always skip the current frame, additionally skip as many frames as
    # provided by skip_frames
    next_frame = current_frame
    while skip_frames > 0:
        if next_frame is None:
            raise ValueError(f"cannot skip {skip_frames} frames")
        next_frame = next_frame.f_back
        skip_frames -= 1

    prev_tb = None
    while next_frame is not None:
        tb = types.TracebackType(
            tb_next=None,
            tb_frame=next_frame,
            tb_lasti=next_frame.f_lasti,
            tb_lineno=next_frame.f_lineno)
        if prev_tb is not None:
            tb.tb_next = prev_tb
        prev_tb = tb
        next_frame = next_frame.f_back
    return prev_tb



"""
Transaction handling aka THD.
"""

import inspect
import typing
from typing import Any, Awaitable, Callable, Coroutine, Self

from ryz import log
from ryz.core import Err, panic
from ryz.types import T

_RollbackFnAndPreResult = tuple[
    Callable[[Any], Awaitable[None] | None],
    Any,
]

class Thd:
    # TODO:
    #       maybe add modes:
    #           "standard" => execute right away,
    #           "defer" => execute only on commit
    def __init__(self):
        self._is_queue_locked = False
        self._rollback_stack: list[_RollbackFnAndPreResult] = []

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        err_type,
        err_val: BaseException | None,
        err_traceback,
    ):
        self._is_queue_locked = True
        if err_val:
            while len(self._rollback_stack) != 0:
                fn, preresult = self._rollback_stack.pop()
                if inspect.iscoroutine(fn):
                    panic(f"expected corofn, but coroutine {fn}")
                try:
                    if inspect.iscoroutinefunction(fn):
                        await fn(preresult)
                        return
                    fn(preresult)
                except Exception as err:
                    log.warn(
                        "catch err (below) during rollback, during execution"
                        f" of fn {fn} => continue",
                    )
                    log.catch(err)

    def a(
        self,
        fn: Callable[[], T],
        rollback_fn: Callable[[T], Any],
    ) -> T:
        if self._is_queue_locked:
            panic("thd queue is locked")
        f = fn()
        self._rollback_stack.append((rollback_fn, f))
        return f

    def a_delete(
        self,
        fn: Callable[[], T],
    ) -> T:
        return self.a(fn, lambda d: getattr(d, "delete")())

    def a_delete_arr_index(
        self,
        index: int,
        fn: Callable[[], T],
    ) -> T:
        return self.a(
            fn,
            lambda arr: typing.cast(list, arr)[index].delete(),
        )

    async def aa_delete(
        self,
        fn: Coroutine[Any, Any, T],
    ) -> T:
        async def delete(val: T):
            getattr(val, "delete").delete()

        return await self.aa(fn, delete)

    async def aa(
        self,
        fn: Coroutine[Any, Any, T],
        rollback_corofn: Callable[[T], Awaitable[Any]],
    ) -> T:
        if self._is_queue_locked:
            Err("thd queue")
        f = await fn
        self._rollback_stack.append((rollback_corofn, f))
        return f




from typing import Generic, Iterable, TypeVar

T = TypeVar("T")
class Range(Generic[T]):
    """
    Represents range between min and max values.

    Min and Max are always both inclusive.
    """
    def __init__(self, min_: T, max_: T):
        self.min = min_
        self.max = max_

    def get_python_range(self) -> Iterable[T]:
        return range(self.min, self.max + 1)  # type: ignore

    def contains(self, val: T) -> bool:
        return self.min <= val and val <= self.max  # type: ignore



import re
from typing import Any

from ryz.cls import Static


class StringUtils(Static):
    @staticmethod
    def stringify(value: dict, separator: str = ", ") -> str:
        """
        Transforms dictionary into string representation.
        """
        option_strs: list[str] = []

        for k, v in value.items():
            option_strs.append(f"{k}={v}")

        return separator.join(option_strs)

    @staticmethod
    def get_titled_value(
        title: str,
        value: Any | None = None,
    ) -> str:
        titled_value: str = title
        if value:
            titled_value = f"{title} <{value!s}>"

        return titled_value

    @staticmethod
    def has_cyrillic(text: str) -> bool:
        """
        Checks if text contains any Cyrillic characters.
        """
        return bool(re.search("[а-яА-Я]", text))

    @staticmethod
    def remove_non_alpha(s: str) -> str:
        """
        Remove all non-alpha characters from string.
        """
        # https://stackoverflow.com/a/22521156
        regex = re.compile("[^a-zA-Z]")
        return regex.sub("", s)

    @staticmethod
    def remove_non_alnum(s: str) -> str:
        """
        Remove all non-alpha characters from string.
        """
        # https://stackoverflow.com/a/22521156
        regex = re.compile("[^a-zA-Z0-9]")
        return regex.sub("", s)



"""
Smart data containers.
"""

from typing import Generic, TypeVar

from ryz.core import Err, Ok, Res, ecode
from ryz.range import Range

T = TypeVar("T")

class Keeper(Generic[T]):
    """
    Manages acquisition strategy for some set of values of the same type.

    In default implementations is supposed to give unique value each new
    request, and to avoid overflows, supports freeing methods, so previously
    obtained values can be returned to the keeper.
    """
    def recv(self) -> Res[T]:
        """
        Receive a new unique val from the keeper.
        """
        raise NotImplementedError

    def free(self, val: T) -> Res[None]:
        """
        Frees val so it's again available in the container.
        """
        raise NotImplementedError

class IntKeeper(Keeper[int]):
    """
    Holds available ints for things like consequent ids.
    """
    def __init__(self, range_: Range[int] = Range(0, 1_000_000)) -> None:
        super().__init__()
        self._range = range_
        self._given: set[int] = set()

    def recv(self) -> Res[int]:
        for possible in self._range.get_python_range():
            if possible not in self._given:
                self._given.add(possible)
                return Ok(possible)
        return Err("no available values")

    def free(self, val: int) -> Res[None]:
        if val not in self._given:
            return Err(f"val {val}", ecode.NotFound)
        self._given.remove(val)
        return Ok(None)



"""
Tools for working with dict-like objects.
"""
from typing import TypeVar

from ryz.core import Err, Ok, Res, ecode

T = TypeVar("T")

def get_recursive(d: dict, key: str, default: T | None = None) -> Res[T]:
    for k, v in d.items():
        if key == k:
            return Ok(v)
        if isinstance(v, dict):
            nested_res = get_recursive(v, key)
            if (
                isinstance(nested_res, Err)
                and nested_res.err.code == ecode.NotFound
            ):
                continue
            return nested_res
    if default is None:
        return Err(f"val for key {key}", ecode.NotFound)
    return Ok(default)



from __future__ import annotations

from typing import Any, TypeVar

SingletonInstance = TypeVar("SingletonInstance")

class SingletonMeta(type):
    """Singleton metaclass for implementing singleton patterns.

    See:
        https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """
    __instances: dict[type, object] = {}

    def __call__(cls, *args, **kwargs) -> Any:
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)
        return cls.__instances[cls]

    def try_discard(cls) -> bool:
        try:
            del cls.__instances[cls]
        except KeyError:
            return False
        return True

class Singleton(metaclass=SingletonMeta):
    """Singleton base class."""
    @classmethod
    def ie(cls: type[SingletonInstance]) -> SingletonInstance:
        """Gets the single instance of the Singleton.

        Returns:
            Instance the Singleton holds.
        """
        return cls()



import time

timedata = time.struct_time

def utc() -> float:
    return time.time()

def delta(d: float, f: float | None = None) -> float:
    """
    Calculates delta timestamp from "f" (or current moment) adding given delta.
    """
    f = f if f is not None else utc()
    return f + d

def local() -> timedata:
    return time.localtime()

def fmt(f: str, td: timedata | None = None):
    return time.strftime(f, td if td is not None else time.gmtime())



def get_fqname(obj: object):
    # ref: https://stackoverflow.com/a/2020083/14748231
    klass = obj.__class__
    module = klass.__module__
    if (
            module is None
            # avoid outputs like "builtins.str"
            or module == "builtins"):
        return klass.__qualname__
    return module + "." + klass.__qualname__



from typing import Generic

from pydantic.generics import GenericModel

from ryz.types import T


class ptr(GenericModel, Generic[T]):
    """
    Points to some target.
    """
    target: T

    class Config:
        arbitrary_types_allowed = True



from typing import Generic, Protocol, TypeVar

T_co = TypeVar("T_co", covariant=True)
class ArbFn(Protocol, Generic[T_co]):
    def __call__(self, *args, **kwargs) -> T_co: ...

class FnSpec(Generic[T_co]):
    """
    Holds information about some function.
    """
    def __init__(
        self,
        func: ArbFn[T_co],
        args: tuple | None = None,
        kwargs: dict | None = None,
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def call(
        self,
        *,
        prepended_extra_args: tuple | None = None,
        appended_extra_args: tuple | None = None,
        prepended_extra_kwargs: dict | None = None,
        appended_extra_kwargs: dict | None = None,
    ) -> T_co:
        """
        Calls a specified in this spec function.
        """
        args: tuple = ()
        if self.args is not None:
            args = self.args

        kwargs: dict = {}
        if self.kwargs is not None:
            kwargs = self.kwargs

        if prepended_extra_args is None:
            prepended_extra_args = ()
        if appended_extra_args is None:
            appended_extra_args = ()
        if prepended_extra_kwargs is None:
            prepended_extra_kwargs = {}
        if appended_extra_kwargs is None:
            appended_extra_kwargs = {}

        final_args: tuple = prepended_extra_args + args + appended_extra_args
        final_kwargs: dict = {
            **prepended_extra_kwargs,
            **kwargs,
            **appended_extra_kwargs,
        }

        return self.func(*final_args, **final_kwargs)



import subprocess
from typing import Callable

from ryz import log


class SubprocUtils:
    @classmethod
    def popen(
        cls,
        cmd: str,
        fn: Callable[[subprocess.Popen], None],
        *,
        must_raise_retcode_err: bool = True,
    ):
        """
        Creates popen and calls given fn with it.

        Fn to read stdout example:
        ```python
        def myfn(proc):
            if proc.stdout:
                for line in proc.stdout:
                    print(line, end="")  # do whatever you want
        ```
        """
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            text=True,
            shell=True,  # noqa: S602
        ) as process:
            try:
                fn(process)
            except KeyboardInterrupt:
                log.info("receive keyboard interrupt => kill subproc")
                process.kill()
                raise
        if must_raise_retcode_err and process.returncode != 0:
            raise subprocess.CalledProcessError(
                process.returncode,
                process.args,
            )




from typing import Callable, Generic, Self

from ryz.types import T


class TreeNode(Generic[T]):
    def __init__(self, val: T, childs: list[Self]):
        self.val = val
        self.childs = childs

    def __str__(self) -> str:
        return f"<TreeNode val = {self.val}, childs = {self.childs}>"

class ReversedTreeNode(Generic[T]):
    def __init__(self, val: T, parent: Self | None = None):
        self.val = val
        self.parent = parent

    def __str__(self) -> str:
        return f"<ReversedTreeNode val = {self.val}, parent = {self.parent}>"

class TreeUtils:
    @classmethod
    async def reverse(
        cls,
        root_node: TreeNode[T],
        *,
        _parent_rnode: ReversedTreeNode[T] | None = None,
    ) -> list[ReversedTreeNode[T]]:
        root_rnode = ReversedTreeNode(root_node.val, _parent_rnode)
        if not root_node.childs:
            return [root_rnode]

        f = []
        for child in root_node.childs:
            f.extend(await cls.reverse(child, _parent_rnode=root_rnode))

        return f

    @classmethod
    async def print(
        cls,
        root_node: TreeNode,
        print_action: Callable[[str], None] = print,
    ):
        def uncover(node: TreeNode, depth: int) -> str:
            msg = ""
            tabs = "\t" * depth

            for c in node.childs:
                msg += f"{tabs} - {c}\n"
                msg += uncover(c, depth + 1)

            return msg

        msg = f"x {root_node}\n"
        depth = 1

        msg += uncover(root_node, depth)
        print_action(msg.strip())




"""
Cryptographic operations.

For password hashing uses bcrypt, which has the salt saved into the hash
itself.

See password hashing https://stackoverflow.com/a/23768422/14748231
"""
import bcrypt


class CryptoUtils:
    @classmethod
    def hash_password(
        cls,
        plain_password_bytes: bytes,
    ) -> bytes:
        return bcrypt.hashpw(
            plain_password_bytes,
            bcrypt.gensalt(12),
        )

    @classmethod
    def check_password(
        cls,
        plain_password_bytes: bytes,
        hashed_password_bytes: bytes,
    ) -> bool:
        return bcrypt.checkpw(
            plain_password_bytes,
            hashed_password_bytes,
        )



import asyncio

from ryz.uuid import uuid4


class Lock:
    def __init__(self) -> None:
        self._evt = asyncio.Event()
        self._evt.set()
        self._owner_token: str | None = None

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, *args):
        assert self._owner_token is not None
        await self.release(self._owner_token)

    def is_locked(self) -> bool:
        return not self._evt.is_set()

    async def acquire(self) -> str:
        await self._evt.wait()
        self._evt.clear()
        self._owner_token = uuid4()
        return self._owner_token

    async def release(self, token: str):
        if self._owner_token is not None and token != self._owner_token:
            raise ValueError("invalid token to unlock")
        self._evt.set()
        self._owner_token = None

    async def wait(self):
        return await self._evt.wait()



"""
Core things we use to maintain python programs.
"""
import re
from inspect import isfunction
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    Literal,
    NoReturn,
    ParamSpec,
    Self,
    TypeAlias,
    TypeVar,
)

from pydantic import BaseModel

from ryz import log, traceback
from ryz.lock import Lock
from ryz.obj import get_fqname

__all__ = [
    "Ok",
    "Res",
    "Err",
    "ecode",
    "resultify",
    "aresultify",
    "secure",
    "asecure",
]

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)  # Success type
U = TypeVar("U")
F = TypeVar("F")
P = ParamSpec("P")
R = TypeVar("R")
TBE = TypeVar("TBE", bound=BaseException)

class ecode:
    Err = "err"
    Panic = "panic_err"
    Val = "val_err"
    NotFound = "not_found_err"
    AlreadyProcessed = "already_processed_err"
    Unsupported = "unsupported_err"
    Lock = "lock_err"

class Err(Exception):
    def __init__(
        self,
        msg: str | None = None,
        code: str = ecode.Err,
        *,
        skip_frames: int = 0,
    ) -> None:
        if not re.match(r"^[a-z][0-9a-z]*(_[0-9a-z]+)*$", code):
            panic(f"invalid code {code}")
        if skip_frames < 0:
            panic(f"`skip_frames` must be positive, got {skip_frames}")
        self.code = code
        self.msg = msg
        final = code
        if msg:
            final += ": " + msg
        # since we don't raise, for each err we create traceback dynamically
        # upon creation, and skip this function frame, as well as others,
        # if the caller's code need it
        traceback.set(self, 1 + skip_frames)
        super().__init__(final)

    def __hash__(self) -> int:
        return hash(self.code)

    def is_(self, code: str) -> bool:
        return self.code == code

    def is_any(self, *code: str) -> bool:
        return self.code in code

    @classmethod
    def from_native(cls, exc: Exception) -> Self:
        return cls("; ".join(exc.args), skip_frames=1)

    def is_ok(self) -> Literal[False]:
        return False

    def is_err(self) -> Literal[True]:
        return True

    @property
    def ok(self) -> None:
        """
        Return `None`.
        """
        return

    @property
    def err(self) -> Self:
        """
        Return the error.
        """
        return self

    def unwrap(self) -> NoReturn:
        """
        Raises an `UnwrapErr`.
        """
        raise self

    def inspect(self, fn: Callable[[T_co], Any]) -> "Res[T_co]":
        """
        Calls a function with the contained value if `Ok`. Returns the original
        result.
        """
        return self

    def ignore(self):
        """
        Used to signify that the result intentially ignored.

        Useful to avoid linter errors on intentional behaviour.
        """
        _ignore(self)

    def track(self, msg: Any = "tracked", v: int = 1) -> str | None:
        if isinstance(self, Exception):
            return log.track(self, msg, v)
        return None

    async def atrack(self, msg: Any = "tracked", v: int = 1) -> str | None:
        if isinstance(self, Exception):
            return await log.atrack(self, msg, v)
        return None

class Ok(Generic[T_co]):
    """
    A value that indicates success and which stores arbitrary data for the
    return value.
    """
    def __init__(self, value: T_co = None) -> None:
        self._value = value

    def __repr__(self) -> str:
        return f"Ok({self._value!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Ok) and self._value == other._value

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        return hash((True, self._value))

    def is_ok(self) -> Literal[True]:
        return True

    def is_err(self) -> Literal[False]:
        return False

    @property
    def ok(self) -> T_co:
        """
        Return the value.
        """
        return self._value

    @property
    def err(self) -> None:
        """
        Return `None`.
        """
        return

    def expect(self, _message: str) -> T_co:
        """
        Return the value.
        """
        return self._value

    def unwrap(self) -> T_co:
        """
        Return the value.
        """
        return self._value

    def inspect(self, fn: Callable[[T_co], Any]) -> "Res[T_co]":
        """
        Calls a function with the contained value if `Ok`. Returns the original
        result.
        """
        fn(self._value)
        return self

    def ignore(self):
        """
        Used to signify that the result intentially ignored.

        Useful to avoid linter errors on intentional behaviour.
        """
        _ignore(self)

    def track(self, msg: Any = "tracked"):
        return

    async def atrack(self, msg: Any = "tracked"):
        return

Res: TypeAlias = Ok[T_co] | Err

CODE_MAX_LEN: int = 256

T = TypeVar("T")
class Coded(BaseModel, Generic[T]):
    """
    Arbitrary data coupled with identification code.

    Useful when data is type that doesn't support ``code() -> str`` signature.
    """
    code: str
    val: T

class Code:
    """
    Manages attached to various objects str codes.
    """
    _code_to_type: dict[str, type] = {}
    _codes: list[str] = []
    _lock: Lock = Lock()

    @classmethod
    def has_code(cls, code: str) -> bool:
        return code in cls._codes

    @classmethod
    async def get_regd_code_by_id(cls, id: int) -> Res[str]:
        await cls._lock.wait()
        if id > len(cls._codes) - 1:
            return Err(f"codeid {id} is not regd")
        return Ok(cls._codes[id])

    @classmethod
    async def get_regd_codeid_by_type(cls, t: type) -> Res[int]:
        code_res = await cls.get_regd_code_by_type(t)
        if isinstance(code_res, Err):
            return code_res
        code = code_res.ok
        return await cls.get_regd_codeid(code)

    @classmethod
    async def get_regd_codes(cls) -> Res[list[str]]:
        await cls._lock.wait()
        return Ok(cls._codes.copy())

    @classmethod
    async def get_regd_code_by_type(cls, t: type) -> Res[str]:
        await cls._lock.wait()
        for c, t_ in cls._code_to_type.items():
            if t_ is t:
                return Ok(c)
        return Err(f"type {t} is not regd")

    @classmethod
    async def get_regd_codeid(cls, code: str) -> Res[int]:
        await cls._lock.wait()
        if code not in cls._codes:
            return Err(f"code {code} is not regd")
        return Ok(cls._codes.index(code))

    @classmethod
    async def get_regd_type_by_code(cls, code: str) -> Res[type]:
        await cls._lock.wait()
        if code not in cls._code_to_type:
            return Err(f"code {code} is not regd")
        return Ok(cls._code_to_type[code])

    @classmethod
    async def upd(
        cls,
        types: Iterable[type | Coded[type]],
        order: list[str] | None = None,
    ) -> Res[None]:
        async with cls._lock:
            for t in types:
                final_t: type
                if isinstance(t, Coded):
                    code = t.code
                    final_t = t.val
                else:
                    code_res = cls.get_from_type(t)
                    if isinstance(code_res, Err):
                        log.err(
                            f"cannot get code for type {t}: {code_res.err}"
                            " => skip")
                        continue
                    code = code_res.ok
                    final_t = t

                validate_res = cls.validate(code)
                if isinstance(validate_res, Err):
                    log.err(
                        f"code {code} is not valid:"
                        f" {validate_res.err} => skip")
                    continue

                cls._code_to_type[code] = final_t

            cls._codes = list(cls._code_to_type.keys())
            if order:
                order_res = cls._order(order)
                if isinstance(order_res, Err):
                    return order_res

            return Ok(None)

    @classmethod
    def destroy(cls):
        cls._code_to_type.clear()
        cls._codes.clear()
        cls._lock = Lock()

    @classmethod
    def _order(cls, order: list[str]) -> Res[None]:
        sorted_codes: list[str] = []
        for o in order:
            if o not in cls._codes:
                log.warn(f"unrecornized order code {o} => skip")
                continue
            cls._codes.remove(o)
            sorted_codes.append(o)

        # bring rest of the codes
        sorted_codes.extend(cls._codes)

        cls._codes = sorted_codes
        return Ok(None)

    @classmethod
    def validate(cls, code: str) -> Res[None]:
        if not isinstance(code, str):
            return Err(f"code {code} must be str")
        if code == "":
            return Err("empty code")
        for i, c in enumerate(code):
            if i == 0 and not c.isalpha():
                return Err(
                    f"code {code} must start with alpha")
            if not c.isalnum() and c != "_" and c != ":":
                return Err(
                    f"code {code} can contain only alnum"
                    " characters, underscores or semicolons")
        if len(code) > CODE_MAX_LEN:
            return Err(f"code {code} exceeds maxlen {CODE_MAX_LEN}")
        return Ok(None)

    @classmethod
    def get_from_type(cls, t: type) -> Res[str]:
        if isinstance(t, Coded):
            code = t.code
        else:
            codefn = getattr(t, "code", None)
            if codefn is None:
                return Err(
                    f"msg data {t} must define \"code() -> str\" method")
            if not isfunction(codefn):
                return Err(
                    f"msg data {t} \"code\" attribute must be function,"
                    f" got {codefn}")
            try:
                code = codefn()
            except Exception as err:
                log.catch(err)
                return Err(
                    f"err {get_fqname(err)} occured during"
                    f" msg data {t} {codefn} method call #~stacktrace")

        validate_res = cls.validate(code)
        if isinstance(validate_res, Err):
            return validate_res

        return Ok(code)

def resultify(
    fn: Callable[[], T_co],
    *errs: type[Exception],
) -> Res[T_co]:
    """
    Calls a func and wraps retval to Res - to Err on thrown exception, Ok
    otherwise.

    Useful to integrate non-result functions.
    """
    try:
        res = fn()
    except errs as err:
        return Err.from_native(err)
    return Ok(res)

async def aresultify(
    coro: Coroutine[Any, Any, T_co],
    *errs: type[Exception],
) -> Res[T_co]:
    """
    Calls a func and wraps retval to Res - to Err on thrown exception, Ok
    otherwise.

    Useful to integrate non-result functions.
    """
    try:
        res = await coro
    except errs as err:
        return Err.from_native(err)
    return Ok(res)

def _ignore(res: Res):
    """
    Used to signify that the result intentially ignored.

    Useful to avoid linter errors on intentional behaviour.
    """

def secure(fn: Callable[[], Res[T_co]]) -> Res[T_co]:
    """
    Wraps function raised error into Err(e), or returns as it is.
    """
    try:
        return fn()
    except Exception as err:
        return Err.from_native(err)

async def asecure(coro: Coroutine[Any, Any, Res[T_co]]) -> Res[T_co]:
    """
    Wraps function raised error into Err(e), or returns as it is.
    """
    try:
        return await coro
    except Exception as err:
        return Err.from_native(err)

def panic(msg: str | None = None) -> NoReturn:
    raise Err(msg, ecode.Panic)



import importlib.metadata as _import_meta

# We don't intend to export modules here since it would be cleaner due to
# different modules to have them all expose themselves.

__version__ = _import_meta.version("ryz")



import asyncio
import sys
from multiprocessing import Pipe, Process
from typing import Any, Literal, Protocol

from ryz import log
from ryz.core import Err, Ok, Res, ecode

if sys.platform == "win32":
    from multiprocessing.connection import PipeConnection as PipeConn
else:
    from multiprocessing.connection import Connection as PipeConn


class ProcTarget(Protocol):
    def __call__(self, **kwargs: Any) -> Any: ...

class ProcGroup:
    """
    Organizes processes.

    By default, processes are using Pipes to communicate.

    Args:
        max_procs:
            Maximum processes to handle. Defaults to -1, which is unlimited.
    """

    def __init__(self, max_procs: int = -1) -> None:
        self._procs: dict[int, tuple[Process, PipeConn]] = {}
        self._key_to_pid: dict[str, int] = {}
        self._max_procs = max_procs
        self.proc_dereg_method: Literal["kill", "terminate"] = "terminate"

    def has(self, pid: int) -> bool:
        return pid in self._procs

    def has_key(self, key: str) -> bool:
        return key in self._key_to_pid

    def _can_reg_by_limit(self) -> bool:
        return \
            self._max_procs < 0 or (len(self._procs) + 1 <= self._max_procs)

    def reg(
            self,
            target: ProcTarget,
            key: str | None = None,
            *,
            proc_kwargs: dict[str, Any] | None = None) -> Res[int]:
        """
        Regs a new process.

        Process's target can only accept kwargs, args are reserved for
        interfacing uses.
        """
        if not self._can_reg_by_limit():
            return Err(
                "cannot reg a new process:"
                f" limit {self._max_procs} is exceeded")

        parent_pipe, child_pipe = Pipe()
        proc = Process(
            target=target,
            args=(child_pipe,),
            kwargs=proc_kwargs if proc_kwargs else {})
        proc.start()

        if proc.pid is None:
            return Err(
                f"{proc} has been started, but the pid is unassigned",
            )
        if self.has(proc.pid):
            proc.kill()
            return Err(
                "new process is started with the same pid as regd"
                " one => kill new process")

        if key:
            if key in self._key_to_pid:
                return Err(f"key {key} is already regd")
            self._key_to_pid[key] = proc.pid

        self._procs[proc.pid] = (proc, parent_pipe)
        return Ok(proc.pid)

    def get_pid_by_key(self, key: str) -> Res[int]:
        if key not in self._key_to_pid:
            return Err(f"key {key}", ecode.NotFound)
        return Ok(self._key_to_pid[key])

    def try_dereg_key(self, key: str) -> Res[bool]:
        pid_res = self.get_pid_by_key(key)
        if isinstance(pid_res, Err):
            return pid_res
        return self.try_dereg(pid_res.unwrap())

    def _end_proc(self, proc: Process):
        if self.proc_dereg_method == "kill":
            proc.kill()
        elif self.proc_dereg_method == "terminate":
            proc.terminate()
        else:
            log.err(
                f"unrecoznized {self.proc_dereg_method}"
                " => use \"terminate\"")

    def try_dereg(self, pid: int) -> Res[bool]:
        """
        Deregs a process by pid.

        The process is killed if it's still alive.
        """
        if pid not in self._procs:
            return Ok(False)

        proc, _ = self._procs[pid]
        if proc.is_alive():
            self._end_proc(proc)

        for key, map_pid in self._key_to_pid.items():
            if map_pid == pid:
                del self._key_to_pid[key]
                break

        del self._procs[pid]
        return Ok(True)

    def recv(self, pid: int) -> Res[Any]:
        proc_data = self._get_proc(pid)
        if isinstance(proc_data, Err):
            return proc_data
        _, pipe = self._get_proc(pid).unwrap()
        data = pipe.recv()
        return Ok(data)

    def recv_key(self, key: str) -> Res[Any]:
        pid_res = self.get_pid_by_key(key)
        if isinstance(pid_res, Err):
            return pid_res
        return self.recv(pid_res.unwrap())

    async def async_recv_key(self, key: str, period: float = 1.0) -> Res[Any]:
        pid_res = self.get_pid_by_key(key)
        if isinstance(pid_res, Err):
            return pid_res
        return await self.async_recv(pid_res.unwrap(), period)

    async def async_recv(self, pid: int, period: float = 1.0) -> Res[Any]:
        """
        Same as recv(), but async.

        Due to problems with implementing this both for Linux and Windows,
        a simple periodic pipe.poll() + pipe.recv() is used. Due to that,
        providing the "period" var is required.
        """
        _, pipe = self._get_proc(pid).unwrap()
        while True:
            if pipe.poll():
                return Ok(pipe.recv())
            await asyncio.sleep(period)

    def send(self, pid: int, data: Any) -> Res[None]:
        proc_data = self._get_proc(pid)
        if isinstance(proc_data, Err):
            return proc_data
        proc_res = self._get_proc(pid)
        if isinstance(proc_res, Err):
            return proc_res
        _, pipe = proc_res.unwrap()
        pipe.send(data)
        return Ok(None)

    def send_key(self, key: str, data: Any) -> Res[None]:
        pid_res = self.get_pid_by_key(key)
        if isinstance(pid_res, Err):
            return pid_res
        pid = pid_res.unwrap()
        return self.send(pid, data)

    def _get_proc(self, pid: int) -> Res[tuple[Process, PipeConn]]:
        if not self.has(pid):
            return Err(f"proc with pid {pid}", ecode.NotFound)
        proc, pipe = self._procs[pid]
        if not proc.is_alive():
            self.try_dereg(pid).unwrap()
            return Err("process is closed")
        return Ok((proc, pipe))



import sys
import tempfile
import traceback
import typing
from pathlib import Path
from typing import Any, NoReturn

from aiofile import async_open
from loguru import logger as _logger

from ryz.obj import get_fqname
from ryz.uuid import uuid4

err_track_dir: Path = Path(tempfile.gettempdir(), "ryz_err_track_dir")
is_debug: bool = False
std_verbosity: int = 1
"""
Verbosity level for stdout/stderr.

For all other targets verbosity is not applied - all msgs are passed to
the sink as it is (yet then it can be blocked there according to the
sink's configuration).

Levels:
    0. silent
    1. cozy chatter
    2. rap god

Methods that produce logging accept variable "v" which defines the
minimal level of verbosity required to make the intended log. For example
if "info('hello', v=1)", the info message would only be produced on
verbosity level 1 or 2.

For debug logs verbosity level is unavailable - they must be emitted
always for their level.
"""

def debug(*args, sep: str = ", "):
    if is_debug:
        _logger.debug(sep.join([str(arg) for arg in args]))

def info(msg: Any, v: int = 1):
    if v < 1:
        return
    if std_verbosity >= v:
        _logger.info(msg)

def warn(msg: Any, v: int = 1):
    if v < 1:
        return
    if std_verbosity >= v:
        _logger.warning(msg)

def err(msg: Any, v: int = 1):
    if v < 1:
        return
    if std_verbosity >= v:
        _logger.error(msg)

def catch(err: Exception, v: int = 1):
    if v < 1:
        return
    if std_verbosity >= v:
        _logger.exception(err)

def err_or_catch(
    err_: Exception, catch_if_v_equal_or_more: int,
):
    if std_verbosity >= catch_if_v_equal_or_more:
        catch(err_)
        return
    err(err)

def fatal(msg: Any, *, exit_code: int = 1) -> NoReturn:
    err(f"FATAL({exit_code}) :: {msg}")
    sys.exit(exit_code)

def _try_get_err_traceback_str(err: Exception) -> str | None:
    """
    Copy of err_utils.try_get_traceback_str to avoid circulars.
    """
    s = None
    tb = err.__traceback__
    if tb:
        extracted_list = traceback.extract_tb(tb)
        s = ""
        for item in traceback.StackSummary.from_list(
                extracted_list).format():
            s += item
    return s

def _get_track_data(
    err_: Exception,
    msg: Any,
    v: int = 1,
) -> tuple[str, Path, str, str]:
    msg = str(msg)
    err_track_dir.mkdir(parents=True, exist_ok=True)
    sid = uuid4()
    track_path = Path(err_track_dir, f"{sid}.log")

    err_ = typing.cast(Exception, err_)
    file_content = _try_get_err_traceback_str(err_)

    if not file_content:
        file_content = ""
    # for filled file_content, append newline operator to separate
    # traceback from err dscr
    elif not file_content.endswith("\n"):
        file_content += "\n"

    # add err dscr
    err_msg = _get_msg(err_)
    err_dscr = get_fqname(err_)
    if err_msg:
        err_dscr += ": " + err_msg
    file_content += err_dscr
    final_msg = msg + f"; {err_dscr}" + f"; $track::{track_path}"

    return sid, track_path, file_content, final_msg

def track(
    err_: Exception,
    msg: Any = "tracked",
    v: int = 1,
) -> str | None:
    """
    Tracks an err with attached msg.

    The err traceback is written to <log.err_track_dir>/<sid>.log, and the
    msg is logged with the sid. This allows to find out error's traceback
    in a separate file by the original log message.

    If ``v`` parameter doesn't match current verbosity, nothing will be
    done.

    If provided err object is Trackable, the complete track file content
    will be retrieved from err.get_track_file_content().

    Returns tracksid or None.
    """
    if std_verbosity < v:
        return None

    sid, track_path, file_content, final_msg = _get_track_data(
        err_, msg, v)

    with track_path.open("w+") as f:
        f.write(file_content)
    err(final_msg, v)
    return sid

async def atrack(
    err_: Exception,
    msg: Any = "tracked",
    v: int = 1,
) -> str | None:
    """
    Asynchronous version of ``log.track``.
    """
    if std_verbosity < v:
        return None

    sid, track_path, file_content, final_msg = _get_track_data(
        err_, msg, v,
    )

    async with async_open(track_path, "w+") as f:
        await f.write(file_content)
    err(final_msg, v)
    return sid

@staticmethod
def _get_msg(err_: Exception) -> str:
    return ", ".join([str(a) for a in err_.args])



"""Operations with strings.

Typically here collected functions which return string.
"""
import re
from typing import Any

from ryz.cls import Static


class FormatUtils(Static):
    @staticmethod
    def format_chain(lst: list[Any]) -> str:
        return " -> ".join([str(x) for x in lst])

    @staticmethod
    def kebabify(name: str) -> str:
        """Converts name to kebab-case.

        Args:
            name:
                Name to convert.

        Returns:
            Name converted.
        """
        # All underscores are simply replaced
        return name.lower().replace("_", "-")

    @staticmethod
    def snakefy(name: str) -> str:
        """Converts name to snake_case.

        Note that this is not reversible using camelfy().

        Args:
            name:
                Name to convert.

        Returns:
            Name converted.
        """
        # Reference: https://stackoverflow.com/a/1176023/14748231
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        name = re.sub("__([A-Z])", r"_\1", name)
        name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
        return name.lower()

    @staticmethod
    def pascalify(name: str) -> str:
        """Converts name to PascalCase.

        Args:
            name:
                Name to convert.

        Returns:
            Name converted.
        """
        result: str = ""
        is_previous_underscore: bool = True

        for i, char in enumerate(name):
            if is_previous_underscore:
                if char == "_":
                    result += char
                else:
                    result += char.upper()
            elif char != "_":
                result += char
            else:
                try:
                    next_char: str = name[i + 1]
                except IndexError:
                    result += char
                else:
                    if next_char == "_":
                        result += char
            is_previous_underscore = char == "_"

        return result



import os

from ryz.core import Err, Ok, Res, ecode


def get(key: str, default: str | None = None) -> Res[str]:
    s = os.environ.get(key, default)
    if s is None:
        return Err(f"cannot find environ {key}", ecode.NotFound)
    return Ok(s)

def get_bool(key: str, default: str | None = None) -> Res[bool]:
    env_val = get(key, default)
    if isinstance(env_val, Err):
        return env_val
    env_val = env_val.ok

    match env_val:
        case "0":
            return Ok(False)
        case "1":
            return Ok(True)
        case _:
            return Err(
                f"{key} expected to be \"1\" or \"0\", but got {env_val}",
            )
